from __future__ import annotations
from loguru import logger
import polars as pl
import pandas as pd
import typing as t
import subprocess
import time
from uptrain.integrations import promptfoo_utils as pr_u

timestr = time.strftime("%m_%d_%Y-%H:%M:%S")  # used for filename

__all__ = ["EvalPromptfoo"]


class EvalPromptfoo:
    def __init__(self) -> None:
        return None

    def install(self):
        try:
            subprocess.run(["npm", "install", "promptfoo@" + pr_u.PROMPTFOO_version])
            logger.success("Install Successfully")
        except Exception as e:
            logger.error("Error installing Promptfoo:" + e)
            logger.error("Try running npx promptfoo@" + pr_u.PROMPTFOO_version)

    def evaluate(
        self,
        evals_list: t.Union[list],
        evals_weight: t.Union[list],
        input_data: t.Union[list[dict], pl.DataFrame, pd.DataFrame],
        prompts: t.Union[str],
        providers: t.Union[list],
        threshold: t.Optional[float] = 0.5,
        output_file: t.Optional[str] = "results_" + timestr + ".csv",
        port: t.Optional[int] = pr_u.generate_open_port(),
        eval_model: t.Optional[str] = "gpt-3.5-turbo-1106",
        project_name: t.Optional[str] = "EvalPromptfoo " + str(timestr),
        kill_process_on_exit: t.Optional[bool] = True,
    ) -> None:
        try:
            evals_compiled = pr_u.compile_evals(evals_list, evals_weight)
            if isinstance(input_data, pl.DataFrame):
                input_data = input_data.to_dicts()
            elif isinstance(input_data, pd.DataFrame):
                input_data = input_data.to_dict(orient="records")
            yaml_data = {
                "prompts": prompts,
                "providers": providers,
                "tests": [
                    {
                        "description": "Data " + str(i + 1),
                        "vars": {
                            "question": input_data[i]["question"],
                            "context": input_data[i]["context"],
                        },
                        "threshold": threshold,
                        "assert": [
                            {
                                "type": "python",
                                "value": pr_u.format_uptrain_template(
                                    input_data[i],
                                    evals_compiled[j],
                                    threshold,
                                    eval_model,
                                    project_name,
                                ),
                            }
                            for j in range(len(evals_compiled))
                        ],
                    }
                    for i in range(len(input_data))
                ],
            }
            pr_u.generate_promptfoo_yaml_file(yaml_data)
            try:
                subprocess.run(
                    [
                        "npx",
                        "--yes",
                        "promptfoo@" + pr_u.PROMPTFOO_version,
                        "eval",
                        "-o",
                        output_file,
                    ]
                )
                logger.success("Evaluations successfully generated")
                logger.success("Saved results to file: " + output_file)
                try:
                    subprocess.run(
                        [
                            "npx",
                            "promptfoo@" + pr_u.PROMPTFOO_version,
                            "view",
                            "-y",
                            "-p",
                            str(port),
                        ]
                    )
                except Exception as e:
                    logger.error(
                        "Something failed: Try running !npx promptfoo@latest view" + e
                    )

            except Exception as e:
                logger.error(f"Evaluation failed with error: {e}")
                raise e
        except Exception:
            if kill_process_on_exit:
                try:
                    pid = subprocess.check_output(["lsof", "-t", "-i:" + str(port)])
                    pids_str = pid.decode()
                    pids = [
                        int(pid)
                        for pid in pids_str.split("\n")
                        if pid.strip().isdigit()
                    ]
                    try:
                        for pid in pids:
                            subprocess.run(["kill", str(pid)])
                        print(
                            f"Closed application at http://localhost:{port} successfully"
                        )
                        print("To reopen it try running view method in EvalPromptfoo")
                    except subprocess.CalledProcessError:
                        pass
                except Exception:
                    print("No running process to close")
            else:
                return None

    def view(
        self, port: t.Optional[int] = pr_u.generate_open_port(),
        kill_process_on_exit: t.Optional[bool] = True,
        ) -> None:
        try:
            try:
                subprocess.run(
                    [
                        "npx",
                        "promptfoo@" + pr_u.PROMPTFOO_version,
                        "view",
                        "-y",
                        "-p",
                        str(port),
                    ]
                )
            except Exception as e:
                logger.error(
                    "Something failed: Try running !npx promptfoo@latest view" + e
                )
        
        except:
            if kill_process_on_exit:
                try:
                    pid = subprocess.check_output(["lsof", "-t", "-i:" + str(port)])
                    pids_str = pid.decode()
                    pids = [
                        int(pid)
                        for pid in pids_str.split("\n")
                        if pid.strip().isdigit()
                    ]
                    try:
                        for pid in pids:
                            subprocess.run(["kill", str(pid)])
                        print(
                            f"Closed application at http://localhost:{port} successfully"
                        )
                        print(f"To reopen it try running view method in EvalPromptfoo")
                    except subprocess.CalledProcessError as e:
                        pass
                except:
                    print("No running process to close")
            else:
                return None

    def custom(self, command: t.Union[str] = "init"):
        try:
            subprocess.run(["npx", "promptfoo@" + pr_u.PROMPTFOO_version, command])
        except Exception as e:
            logger.error("Something failed: " + e)



"""Implements `Check` objects used for LLM evaluation purposes.
"""

from __future__ import annotations
from dataclasses import dataclass
import os
import typing as t

from loguru import logger
import polars as pl
from pydantic import BaseModel

from uptrain.operators.base import (
    Operator,
    TransformOp,
    ColumnOp,
    deserialize_operator,
)
from uptrain.utilities import jsonload, jsondump, to_py_types, clear_directory
from uptrain.framework.base import OperatorDAG, Settings

__all__ = ["Check", "CheckSet", "ExperimentArgs"]


class Check:
    """A simple check that runs the given list of table operators in sequence.

    Attributes:
        name (str): Name of the check.
        operators (list[TableOp]): A list of operators to run in sequence on the input data. The output of each
            operator is passed as input to the next operator.
        plots (list[Operator]): How to plot the output of the check.
    """

    name: str
    operators: list[Operator]
    plots: list[Operator]

    def __init__(
        self,
        name: str,
        operators: list[Operator],
        plots: t.Union[list[Operator], None] = None,
    ):
        self.name = name
        self.operators = operators
        self.plots = plots if plots is not None else []

    def setup(self, settings: "Settings"):
        self._settings = settings

        # no need to add the plot operator to the dag, since it's run later
        self._op_dag = OperatorDAG(name=self.name)
        for i, op in enumerate(self.operators):
            if i == 0:
                deps = []
            else:
                deps = [f"operator_{i-1}"]
            self._op_dag.add_step(f"operator_{i}", op, deps=deps)
        self._op_dag.setup(settings)

        return self

    def run(
        self, data: t.Union[pl.DataFrame, None] = None
    ) -> t.Union[pl.DataFrame, None]:
        """Run this check on the given data."""
        node_inputs = {"operator_0": data}

        if len(self.operators):
            # pick output from the last op in the sequence
            name_final_node = f"operator_{len(self.operators) - 1}"
            node_outputs = self._op_dag.run(
                node_inputs=node_inputs,
                output_nodes=[name_final_node],
            )
            return node_outputs[name_final_node]
        else:
            return data

    def dict(self) -> dict:
        """Serialize this check to a dict."""
        return {
            "name": self.name,
            "operators": [to_py_types(op) for op in self.operators],
            "plots": [to_py_types(op) for op in self.plots],
        }  # serializes only the attributes of the class, like pydantic models

    @classmethod
    def from_dict(cls, data: dict) -> "Check":
        """Deserialize a check from a dict of its parameters."""
        operators = [deserialize_operator(op) for op in data["operators"]]
        plots = [deserialize_operator(op) for op in data["plots"]]
        return cls(name=data["name"], operators=operators, plots=plots)  # type: ignore


class CheckSet:
    """Container for a set of checks to run together. This is the entrypoint to Uptrain for users.

    Attributes:
        source (Operator): The source operator to run. Specifies where to get the data from.
        preprocessors (list[TableOp]): A list of operators to run on the input data before running the checks.
        checks (list[Check]): The set of checks to run on the input data.
    """

    source: Operator
    checks: list[Check]
    preprocessors: list[TransformOp]
    postprocessors: list[Operator]

    def __init__(
        self,
        source: Operator,
        checks: list[t.Any],
        preprocessors: t.Union[list[TransformOp], None] = None,
        postprocessors: t.Union[list[Operator], None] = None,
    ):
        self.source = source
        self.checks = checks
        self.preprocessors = preprocessors if preprocessors is not None else []
        self.postprocessors = postprocessors if postprocessors is not None else []

        # verify all checks have different names
        check_names = [check.name for check in checks]
        assert len(set(check_names)) == len(check_names), "Duplicate check names"
        for check in checks:
            assert isinstance(check, Check), "Each check must be an instance of Check"

    def setup(self, settings: Settings):
        """Create the logs directory, or clear it if it already exists. Also, persist the
        evaluation config.
        """
        self._settings = settings
        logs_dir = self._settings.logs_folder
        if not os.path.exists(logs_dir):
            os.makedirs(logs_dir)
        else:
            clear_directory(logs_dir)

        logger.info(f"Uptrain Logs directory: {logs_dir}")

        # persist the check-set as well as the corresponding settings
        self.serialize(os.path.join(logs_dir, "config.json"))
        self._settings.serialize(os.path.join(logs_dir, "settings.json"))

        self.source.setup(self._settings)
        for preprocessor in self.preprocessors:
            preprocessor.setup(self._settings)
        for check in self.checks:
            check.setup(self._settings)
        for postprocessor in self.postprocessors:
            postprocessor.setup(self._settings)
        return self

    def run(self):
        """Run all checks in this set."""
        from uptrain.operators import JsonWriter

        logger.info("CheckSet Status: Starting checkset")

        source_output = self.source.run()["output"]
        if source_output is None:
            raise RuntimeError("Dataset read from the source is: None")
        if len(source_output) == 0:
            raise RuntimeError("Dataset read from the source is: empty")
        logger.info("CheckSet Status: Dataset loaded from source")

        if len(self.preprocessors) > 0:
            for preprocessor in self.preprocessors:
                source_output = preprocessor.run(source_output)["output"]
                assert source_output is not None, "Output of preprocessor is None"

            # persist the preprocessed input for debugging
            JsonWriter(
                fpath=os.path.join(
                    self._settings.logs_folder, "preprocessed_input.jsonl"
                )
            ).setup(self._settings).run(source_output)

        logger.info("CheckSet Status: Preprocessing Done")

        consolidated_output = {}
        for check in self.checks:
            logger.info(f"CheckSet Status: Check {check.name} Started")
            check_output = check.run(source_output)
            assert check_output is not None, f"Output of check {check.name} is None"
            self._get_sink_for_check(self._settings, check).run(check_output)
            logger.info(f"CheckSet Status: Check {check.name} Completed")

            if len(self.postprocessors):
                if not all(isinstance(op, ColumnOp) for op in check.operators):
                    continue
                for col in check_output.columns:
                    consolidated_output[col] = check_output[col]
        logger.info("CheckSet Status: All Checks Completed")

        if len(self.postprocessors):
            consolidated_output = pl.DataFrame(consolidated_output)
            for postprocessor in self.postprocessors:
                consolidated_output = postprocessor.run(consolidated_output)["output"]
                assert (
                    consolidated_output is not None
                ), "Output of postprocessor is None"

            # persist the postprocessed input for debugging
            JsonWriter(
                fpath=os.path.join(
                    self._settings.logs_folder, "postprocessed_input.jsonl"
                )
            ).setup(self._settings).run(consolidated_output)
        logger.info("CheckSet Status: Postprocessing Done")

    @staticmethod
    def _get_sink_for_check(settings: Settings, check: Check):
        """Get the sink operator for this check."""
        from uptrain.operators import JsonWriter

        fname = check.name.replace(" ", "_") + ".jsonl"
        return JsonWriter(fpath=os.path.join(settings.logs_folder, fname))

    @classmethod
    def from_dict(cls, data: dict) -> "CheckSet":
        checks = [Check.from_dict(check) for check in data.get("checks", [])]
        return cls(
            source=deserialize_operator(data["source"]),
            preprocessors=[
                deserialize_operator(op) for op in data.get("preprocessors", [])
            ],  # type: ignore
            checks=checks,
            postprocessors=[
                deserialize_operator(op) for op in data.get("postprocessors", [])
            ],  # type: ignore
        )

    def dict(self) -> dict:
        return {
            "source": to_py_types(self.source),
            "preprocessors": [to_py_types(op) for op in self.preprocessors],
            "checks": [check.dict() for check in self.checks],
            "postprocessors": [to_py_types(op) for op in self.postprocessors],
        }

    @classmethod
    def deserialize(cls, fpath: str) -> "CheckSet":
        with open(fpath, "r") as f:
            return cls.from_dict(jsonload(f))

    def serialize(self, fpath: t.Optional[str] = None):
        """Serialize this check set along with the run settings to a JSON file."""
        if fpath is None:
            fpath = os.path.join(self._settings.logs_folder, "config.json")

        with open(fpath, "w") as f:
            jsondump(self.dict(), f)


class ExperimentArgs(BaseModel):
    """Container for arguments to run a LLM experiment. This is the entrypoint to run prompt/model tests with Uptrain for users.
    This generates multiple prompts, generate LLM responses for them and parses the response to add multiple columns.

    Attributes:
        prompt_template (str): Prompt template to generate multiple prompts to experiment with.
        prompt_params (dict(str, str)): Dictionary of prompt variables with experiment with.
        models (str or list[str]): Either a single model or list of models to experiment with.
        context_vars (dict or list): List of dataset variables with mapping between variable name in prompt and variable name in dataset
        temperature (float): Temperature for the LLM response generation.
        col_out_response (str): LLM output will be saved under this column name
        col_out_mapping (dict): Mapping used to parse LLM response
    """

    prompt_template: str
    prompt_params: dict = {}
    models: t.Union[str, list[str]]
    context_vars: t.Union[dict, list]
    temperature: float = 1.0
    col_out_response: str = "response"
    col_out_mapping: t.Optional[dict] = None

    def _get_preprocessors(self):
        from uptrain.operators import PromptGenerator, TextCompletion, OutputParser

        "Convert experiment args into checkset preprocessors for execution"
        if isinstance(self.context_vars, list):
            self.context_vars = dict(zip(self.context_vars, self.context_vars))

        preprocessors = [
            PromptGenerator(
                prompt_template=self.prompt_template,
                prompt_params=self.prompt_params,
                models=[self.models] if isinstance(self.models, str) else self.models,
                context_vars=self.context_vars,
            ),
            TextCompletion(
                col_in_prompt="exp_prompt",
                col_in_model="exp_model",
                col_out_completion=self.col_out_response,
                temperature=self.temperature,
            ),
        ]

        if self.col_out_mapping is not None:
            preprocessors.append(
                OutputParser(
                    col_in_response=self.col_out_response,
                    col_out_mapping=self.col_out_mapping,
                )
            )

        return preprocessors

    def _modify_checks(self, checks):
        "Modify checks to add experiment variables into plots"
        for check in checks:
            if check.plots is not None:
                for plot in check.plots:
                    if plot.kind in ["histogram", "bar", "line"]:
                        if "barmode" not in plot.props:
                            plot.props.update({"barmode": "group"})
                        plot.props["color"] = "exp_experiment_id"
        return checks



from __future__ import annotations
from datetime import datetime, timedelta, tzinfo
import dataclasses
import importlib
import importlib.util
import json
import typing as t
import os
import time

from lazy_loader import load as _lazy_load
from loguru import logger
from pydantic import BaseModel
import numpy as np

# import numpy.typing as npt
# import pyarrow as pa
import polars as pl
import datetime


# -----------------------------------------------------------
# utility routines for JSON serialization - parts picked off
# from the openai-evals codebase
# -----------------------------------------------------------


def to_py_types(obj: t.Any) -> t.Any:
    import inspect

    # for nested dataclasses/pydantic models/operators
    if isinstance(obj, datetime.datetime):
        return str(obj)
    if isinstance(obj, dict):
        return {k: to_py_types(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [to_py_types(v) for v in obj]
    elif dataclasses.is_dataclass(obj):
        return to_py_types(dataclasses.asdict(obj))
    elif hasattr(obj, "_uptrain_op_name"):
        if hasattr(obj, "_uptrain_op_custom"):
            if hasattr(obj, "_uptrain_op_custom_source"):
                source = obj._uptrain_op_custom_source
            else:
                source = inspect.getsource(obj.__class__)
            return {
                "op_name": getattr(obj, "_uptrain_op_name"),
                "params": obj.dict(include=set(obj.__fields__)),
                "source": source,
            }
        else:
            return {
                "op_name": getattr(obj, "_uptrain_op_name"),
                "params": obj.dict(include=set(obj.__fields__)),
            }
    elif isinstance(obj, BaseModel):
        return obj.model_dump()

    # for numpy types
    if isinstance(obj, np.integer):
        return int(obj)
    elif isinstance(obj, np.floating):
        return float(obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()

    return obj


class UpTrainEncoder(json.JSONEncoder):
    """Special json encoder for numpy types"""

    def default(self, obj: t.Any) -> str:
        return to_py_types(obj)


def jsondumps(obj: t.Any, **kwargs) -> str:
    return json.dumps(obj, cls=UpTrainEncoder, **kwargs)


def jsondump(obj: t.Any, fp: t.Any, **kwargs) -> None:
    json.dump(obj, fp, cls=UpTrainEncoder, **kwargs)


def jsonloads(s: str, **kwargs) -> t.Any:
    return json.loads(s, **kwargs)


def jsonload(fp: t.Any, **kwargs) -> t.Any:
    return json.load(fp, **kwargs)


# -----------------------------------------------------------
# utility routines for converting between Arrow and Numpy
# -----------------------------------------------------------


# def array_np_to_arrow(arr: npt.NDArray) -> pa.Array:
#     assert arr.ndim in (1, 2), "Only 1D and 2D arrays are supported."
#     if arr.ndim == 1:
#         return pa.array(arr)
#     else:
#         dim1, dim2 = arr.shape
#         return pa.ListArray.from_arrays(
#             np.arange(0, (dim1 + 1) * dim2, dim2), arr.ravel()
#         )


# def array_arrow_to_np(arr: pa.Array) -> npt.NDArray:
#     if isinstance(arr, pa.ChunkedArray):
#         arr = arr.combine_chunks()

#     if not pa.types.is_list(arr.type):
#         return arr.to_numpy(zero_copy_only=False)  # assume a 1D array
#     else:
#         dim1 = len(arr)  # assume a 2D array
#         return np.asarray(arr.values.to_numpy(zero_copy_only=False)).reshape(dim1, -1)


# def arrow_batch_to_table(batch_or_tbl: t.Union[pa.Table, pa.RecordBatch]) -> pa.Table:
#     if not isinstance(batch_or_tbl, pa.Table):
#         return pa.Table.from_batches([batch_or_tbl])
#     else:
#         return batch_or_tbl


# def table_arrow_to_np_arrays(
#     tbl: pa.Table, cols: list[t.Union[str, int]]
# ) -> list[np.ndarray]:
#     return [array_arrow_to_np(tbl.column(c)) for c in cols]


# def np_arrays_to_arrow_table(arrays: list[npt.NDArray], cols: list[str]) -> pa.Table:
#     return pa.Table.from_pydict(
#         {c: array_np_to_arrow(arr) for c, arr in zip(cols, arrays)}
#     )


def polars_to_json_serializable_dict(data: pl.DataFrame):
    data_dictn = data.to_dicts()

    # Remove cases which are not json serializable
    for key in list(data_dictn[0].keys()):
        delete_key = False
        if "datetime" in str(type(data_dictn[0][key])):
            delete_key = True

        if delete_key:
            for row in data_dictn:
                del row[key]

    try:
        json.dumps(data_dictn)
    except Exception:
        for key in list(data_dictn[0].keys()):
            try:
                json.dumps([x[key] for x in data_dictn])
            except Exception:
                for row in data_dictn:
                    del row[key]

    return data_dictn


def polars_to_pandas(data: pl.DataFrame):
    """Convert a polars dataframe to a pandas dataframe"""
    # FIXME: obscure error during pandas conversion through pyarrow for string columns, though
    # they already are utf-8 encoded by arrow/polars?? I tried pandas 1.5.3 and pyarrow>12,
    # as rec-d on github.
    import pandas as pd

    try:
        pd_data = data.to_pandas()
    except Exception:
        # convert to python native types first and then to pandas
        logger.warning(
            "Error converting polars to pandas. Trying to convert to python native types first."
        )

        lazy_load_dep("pyarrow", "pyarrow>=10.0.0")

        # We can't iterate over polars as such, since rust panics can't be caught in python
        # convert to pyarrow first
        all_rows = []
        arrow_data = data.to_arrow()
        for row in range(len(data)):
            try:
                row_dict = {}
                for col in data.columns:
                    row_dict[col] = arrow_data[col][row].as_py()
                all_rows.append(row_dict)
            except Exception as e:
                logger.warning(f"Error converting row {row}: {e}")

        pd_data = pd.DataFrame(all_rows)
    return pd_data


# -----------------------------------------------------------
# routines for working with local files and directories
# -----------------------------------------------------------


def clear_directory(dir_path: str):
    """Clears the directory at dir_path but without deleting the directory itself. `shutil.rmtree` will
    have difficulties with mounted volumes or network drives.
    """
    import shutil

    for filename in os.listdir(dir_path):
        file_path = os.path.join(dir_path, filename)
        if os.path.isfile(file_path) or os.path.islink(file_path):
            os.unlink(file_path)
        elif os.path.isdir(file_path):
            shutil.rmtree(file_path)


# -----------------------------------------------------------
# routines to deal with time
# -----------------------------------------------------------


class Clock:
    """Makes testing easier by anchoring to an older time. This is helpful when
    testing with older datasets. While if initialized with no arguments, the clock
    starts at the current time.
    """

    behind_by: timedelta
    tzone: t.Optional[tzinfo]

    def __init__(self, init_at: datetime):
        tz = init_at.tzinfo
        if tz is None or tz.utcoffset(init_at) is None:
            self.tzone = None
            self.behind_by = datetime.now() - init_at
        else:
            self.tzone = tz
            self.behind_by = datetime.now(tz=tz) - init_at

    def now(self) -> datetime:
        """Return the current time, adjusted by the amount of time the clock is behind."""
        return datetime.now(tz=self.tzone) - self.behind_by

    def sleep(self, seconds: float):
        """If the clock is behind, catch up. Else, sleep for the given duration."""
        seconds_behind = self.behind_by.total_seconds()
        if seconds_behind > 0:
            print(f"advancing the clock by {seconds} seconds")
        else:
            print(f"sleeping for {seconds} seconds")
            time.sleep(seconds)
        self.behind_by = timedelta(seconds=max(seconds_behind - seconds, 0))


class Timer:
    """Context manager for timing code blocks"""

    time: float

    def __enter__(self):
        self.time = time.perf_counter()
        return self

    def __exit__(self, type, value, traceback):
        self.time = round(time.perf_counter() - self.time, 3)


# -----------------------------------------------------------
# routines to deal with optional dependencies
# -----------------------------------------------------------


def lazy_load_dep(import_name: str, package_name: str):
    """Helper function to lazily load optional dependencies. If the dependency is not
    present, the function will raise an error _when used_.

    NOTE: This wrapper adds a warning message at import time.
    """
    try:
        spec = importlib.util.find_spec(import_name)
    except Exception:
        spec = None
    if spec is None:
        logger.warning(
            f"Optional feature dependent on missing package: {import_name} was initialized.\n"
            f"Use `pip install {package_name}` to install the package if running locally."
        )

    return _lazy_load(import_name)



"""
Implement operaores to evaluate response quality i.e. quality of the generated response.
"""

from __future__ import annotations
import typing as t
import json

from loguru import logger
import polars as pl


from uptrain.utilities.prompt_utils import parse_scenario_description

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import register_op, ColumnOp, TYPE_TABLE_OUTPUT
from uptrain.utilities import polars_to_json_serializable_dict
from uptrain.operators.language.llm import LLMMulticlient
from uptrain.operators.language.prompts.output_format import (
    CLASSIFY_JSON_OUTPUT_FORMAT,
    COT_CLASSIFY_JSON_OUTPUT_FORMAT,
)


@register_op
class CustomPromptEvalScore(ColumnOp):
    """
    Evaluate the quality of a response based on a custom prompt.

    Attributes:
    -----------
    prompt: str
        The prompt to evaluate the response quality.
    prompt_var_to_column_mapping: dict[str, str]
        Mapping from variables in the prompt to columns in the data.
    choices: list[str]
        List of choices for the LLM to select from.
    choice_scores: list[float]
        Scores associated for each choice.
    eval_type: str
        Type of evaluation. Options are "classify" and "cot_classify".
    col_out: str
        Column name to store the score.

    """

    prompt: str
    prompt_var_to_column_mapping: dict[str, str]
    choices: list[str]
    choice_scores: list[float]
    eval_type: t.Literal["classify", "cot_classify"] = "cot_classify"
    col_out: str = "score_custom_prompt"

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self.settings = settings
        if self.settings.evaluate_locally and (
            self.settings.uptrain_access_token is None
            or not len(self.settings.uptrain_access_token)
        ):
            self._api_client = LLMMulticlient(settings)
        else:
            self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = polars_to_json_serializable_dict(data)

        try:
            if self.settings.evaluate_locally and (
                self.settings.uptrain_access_token is None
                or not len(self.settings.uptrain_access_token)
            ):
                results = self.evaluate_local(data_send)
            else:
                
                results = self._api_client.evaluate("CustomPromptEval", data_send, {
                    "prompt": self.prompt,
                    "choices": self.choices,
                    "choice_scores": self.choice_scores,
                    "eval_type": self.eval_type,
                    "prompt_var_to_column_mapping": self.prompt_var_to_column_mapping,
                })
        except Exception as e:
            logger.error(f"Failed to run evaluation for `ResponseCompleteness`: {e}")
            raise e

        assert results is not None
        return {
            "output": data.with_columns(
                pl.from_dicts(results).rename({"score_custom_prompt": self.col_out})
            )
        }

    def evaluate_local(self, data):
        """
        Our methodology is based on the model grade evaluation introduced by openai evals.
        """

        input_payloads = []

        for idx, item in enumerate(data):
            subs = item
            subs = {k: item[v] for k, v in self.prompt_var_to_column_mapping.items()}
            grading_prompt_template = self.prompt.format(**subs)
            if self.eval_type == "classify":
                output_vars = {"Choice": "[Your Choice]"}
                grading_prompt_template += CLASSIFY_JSON_OUTPUT_FORMAT.format(
                    num_choices=len(self.choices),
                    choices=self.choices,
                    output_format=json.dumps(output_vars, indent=2),
                )
            elif self.eval_type == "cot_classify":
                output_vars = {
                    "Choice": "[Your Choice]",
                    "Explanation": "[Your Explanation here]",
                }
                grading_prompt_template += COT_CLASSIFY_JSON_OUTPUT_FORMAT.format(
                    num_choices=len(self.choices),
                    choices=self.choices,
                    output_format=json.dumps(output_vars, indent=2),
                )
            input_payloads.append(
                self._api_client.make_payload(idx, grading_prompt_template)
            )
        output_payloads = self._api_client.fetch_responses(
            input_payloads, lambda x: True
        )
        results = []
        for res in output_payloads:
            idx = res.metadata["index"]
            output = {"score_custom_prompt": None}
            try:
                # score_mapping is a mapping from choices to choice_scores
                score_mapping = dict(zip(self.choices, self.choice_scores))
                score = score_mapping[
                    json.loads(res.response.choices[0].message.content)["Choice"]
                ]
                output["score_custom_prompt"] = float(score)
            except Exception:
                logger.error(
                    f"Error when processing payload at index {idx}: {res.error}"
                )
            results.append((idx, output))
        results = [val for _, val in sorted(results, key=lambda x: x[0])]

        return results



"""
This module implements a simple LLM wrapper that can be used to evaluate performance
of LLM applications. 
"""

import typing as t
from datetime import datetime
from loguru import logger
import httpx
import polars as pl
import pandas as pd
import pydantic
import copy
import os
import httpx
from uptrain.operators.base import ColumnOp
from uptrain.utilities.utils import parse_prompt, check_openai_api_key
from uptrain.framework.remote import APIClientWithoutAuth, DataSchema
from uptrain.framework.base import Settings
from uptrain.framework.checks import Check
from uptrain.framework.evals import (
    Evals,
    JailbreakDetection,
    ParametricEval,
    CritiqueTone,
    GuidelineAdherence,
    ResponseMatching,
    ConversationSatisfaction,
)
from uptrain.operators import (
    ResponseFactualScore,
    ContextRelevance,
    ResponseCompleteness,
    ResponseCompletenessWrtContext,
    ResponseConciseness,
    ResponseConsistency,
    ResponseMatchingScore,
    ValidResponseScore,
    JailbreakDetectionScore,
    PromptInjectionScore,
    GuidelineAdherenceScore,
    ConversationSatisfactionScore,
    LanguageCritique,
    ToneCritique,
    ResponseRelevance,
    ContextConciseness,
    ContextReranking,
    SubQueryCompleteness,
    CodeHallucinationScore,
    CustomPromptEvalScore,
    MultiQueryAccuracy,
    ValidQuestionScore
)

from uptrain.framework.rca_templates import RcaTemplate
from uptrain.operators import RagWithCitation

RCA_TEMPLATE_TO_OPERATOR_MAPPING = {RcaTemplate.RAG_WITH_CITATION: RagWithCitation()}

EVAL_TO_OPERATOR_MAPPING = {
    Evals.FACTUAL_ACCURACY: ResponseFactualScore(),
    Evals.CONTEXT_RELEVANCE: ContextRelevance(),
    Evals.CONTEXT_RERANKING: ContextReranking(),
    Evals.CONTEXT_CONCISENESS: ContextConciseness(),
    Evals.RESPONSE_COMPLETENESS: ResponseCompleteness(),
    Evals.RESPONSE_CONCISENESS: ResponseConciseness(),
    Evals.RESPONSE_COMPLETENESS_WRT_CONTEXT: ResponseCompletenessWrtContext(),
    Evals.RESPONSE_CONSISTENCY: ResponseConsistency(),
    Evals.RESPONSE_RELEVANCE: ResponseRelevance(),
    Evals.VALID_RESPONSE: ValidResponseScore(),
    Evals.PROMPT_INJECTION: PromptInjectionScore(),
    Evals.CRITIQUE_LANGUAGE: LanguageCritique(),
    Evals.SUB_QUERY_COMPLETENESS: SubQueryCompleteness(),
    Evals.CODE_HALLUCINATION: CodeHallucinationScore(),
    Evals.MULTI_QUERY_ACCURACY: MultiQueryAccuracy(),
    Evals.QUESTION_COMPLETENESS: ValidQuestionScore(),
}

PARAMETRIC_EVAL_TO_OPERATOR_MAPPING = {
    "JailbreakDetection": JailbreakDetectionScore,
    "GuidelineAdherence": GuidelineAdherenceScore,
    "ConversationSatisfaction": ConversationSatisfactionScore,
    "CritiqueTone": ToneCritique,
    "ResponseMatching": ResponseMatchingScore,
    "CustomPromptEval": CustomPromptEvalScore,
}


class EvalLLM:
    def __init__(self, settings: Settings = None, openai_api_key: str = None) -> None:
        if (openai_api_key is None) and (settings is None):
            raise Exception("Please provide OpenAI API Key")

        if settings is None:
            self.settings = Settings(openai_api_key=openai_api_key)
        else:
            self.settings = settings
        if self.settings.openai_api_key is not None and len(self.settings.openai_api_key):
            response = check_openai_api_key(self.settings.openai_api_key)
            if not response:
                raise ValueError("OpenAI API Key is invalid")

        self.executor = APIClientWithoutAuth(self.settings)

    ####
    def perform_root_cause_analysis(
        self,
        data: t.Union[list[dict], pl.DataFrame, pd.DataFrame],
        rca_template: RcaTemplate,
        scenario_description: t.Union[str, list[str], None] = None,
        schema: t.Union[DataSchema, dict[str, str], None] = None,
        metadata: t.Optional[dict[str, t.Any]] = None,
    ):
        """Perform root cause analysis for the open source user.
        NOTE: This api doesn't log any data.

        Args:
            data: Data to evaluate on. Either a Pandas DataFrame or a list of dicts.
            rca_template: rca template to run.
            schema: Schema of the data. Only required if the data attributes aren't typical (question, response, context).
            metadata: Attributes to attach to this dataset. Useful for filtering and grouping in the UI.

        Returns:
            results: List of dictionaries with each data point and corresponding evaluation results.
        """

        if isinstance(data, pl.DataFrame):
            data = data.to_dicts()
        elif isinstance(data, pd.DataFrame):
            data = data.to_dict(orient="records")

        if schema is None:
            schema = DataSchema()
        elif isinstance(schema, dict):
            schema = DataSchema(**schema)

        if metadata is None:
            metadata = {}

        req_attrs, ser_template = set(), {}
        if rca_template == RcaTemplate.RAG_WITH_CITATION:
            req_attrs.update(
                [schema.question, schema.response, schema.context, schema.cited_context]
            )
        else:
            raise Exception("RCA Template not supported yet")

        dictn = {"scenario_description": scenario_description}
        ser_template.update({"rca_template_name": rca_template.value, **dictn})

        for idx, row in enumerate(data):
            if not req_attrs.issubset(row.keys()):
                raise ValueError(
                    f"Row {idx} is missing required all required attributes for evaluation: {req_attrs}"
                )

        if self.settings.evaluate_locally:
            results = copy.deepcopy(data)
            if rca_template in RCA_TEMPLATE_TO_OPERATOR_MAPPING:
                op = RCA_TEMPLATE_TO_OPERATOR_MAPPING[rca_template]
                op.scenario_description = (
                    scenario_description
                    if not isinstance(scenario_description, list)
                    else scenario_description[idx]
                )
                res = (
                    op.setup(self.settings).run(pl.DataFrame(data))["output"].to_dicts()
                )
            else:
                res = self.evaluate_on_server(data, [ser_template], schema)
            for idx, row in enumerate(res):
                results[idx].update(row)
        else:
            results = self.evaluate_on_server(data, [ser_template], schema)
        return results

    def evaluate(
        self,
        data: t.Union[list[dict], pl.DataFrame, pd.DataFrame],
        checks: list[t.Union[str, Evals, ParametricEval]],
        project_name: str = "Project - " + str(datetime.utcnow()),
        scenario_description: t.Optional[str] = None,
        schema: t.Union[DataSchema, dict[str, str], None] = None,
        metadata: t.Optional[dict[str, str]] = None,
    ):
        """Run an evaluation on the UpTrain server using user's openai keys.
        NOTE: This api doesn't log any data.

        Args:
            project_name: Name of the project.
            data: Data to evaluate on. Either a Pandas DataFrame or a list of dicts.
            checks: List of checks to evaluate on.
            schema: Schema of the data. Only required if the data attributes aren't typical (question, response, context).
            metadata: Attributes to attach to this dataset. Useful for filtering and grouping in the UI.
        Returns:
            results: List of dictionaries with each data point and corresponding evaluation results.
        """

        if isinstance(data, pl.DataFrame):
            data = data.to_dicts()
        elif isinstance(data, pd.DataFrame):
            data = data.to_dict(orient="records")

        if schema is None:
            schema = DataSchema()
        elif isinstance(schema, dict):
            schema = DataSchema(**schema)

        if metadata is None:
            metadata = {}

        checks = [Evals(m) if isinstance(m, str) else m for m in checks]
        for m in checks:
            assert isinstance(m, (Evals, ParametricEval, ColumnOp, list))
            if isinstance(m, list):
                for op in m:
                    assert isinstance(op, ColumnOp)

        req_attrs, ser_checks = set(), []
        for idx, m in enumerate(checks):
            if m in [Evals.SUB_QUERY_COMPLETENESS]:
                req_attrs.update([schema.sub_questions, schema.question])
            elif m in [Evals.CONTEXT_CONCISENESS]:
                req_attrs.update(
                    [schema.question, schema.context, schema.concise_context]
                )
            elif m in [Evals.CONTEXT_RERANKING]:
                req_attrs.update(
                    [schema.question, schema.context, schema.reranked_context]
                )
            elif m in [
                Evals.FACTUAL_ACCURACY,
                Evals.RESPONSE_COMPLETENESS_WRT_CONTEXT,
                Evals.RESPONSE_CONSISTENCY,
                Evals.CODE_HALLUCINATION,
            ]:
                req_attrs.update([schema.question, schema.context, schema.response])
            elif m in [
                Evals.RESPONSE_RELEVANCE,
                Evals.VALID_RESPONSE,
                Evals.RESPONSE_COMPLETENESS,
                Evals.RESPONSE_CONCISENESS,
            ]:
                req_attrs.update([schema.question, schema.response])
            elif m in [Evals.CONTEXT_RELEVANCE]:
                req_attrs.update([schema.question, schema.context])
            elif (
                m in [Evals.CRITIQUE_LANGUAGE]
                or isinstance(m, CritiqueTone)
                or isinstance(m, GuidelineAdherence)
            ):
                req_attrs.update([schema.response])
            elif isinstance(m, ResponseMatching):
                req_attrs.update(
                    [schema.question, schema.response, schema.ground_truth]
                )
            elif isinstance(m, ConversationSatisfaction):
                req_attrs.update([schema.conversation])
            elif m in [Evals.PROMPT_INJECTION] or isinstance(m, JailbreakDetection):
                req_attrs.update([schema.question])

            this_scenario_description = (
                scenario_description
                if not isinstance(scenario_description, list)
                else scenario_description[idx]
            )

            if isinstance(m, ParametricEval):
                dictm = m.model_dump()
                dictm.update({"scenario_description": this_scenario_description})
                ser_checks.append({"check_name": m.__class__.__name__, **dictm})
            elif isinstance(m, Evals):
                dictm = {"scenario_description": this_scenario_description}
                ser_checks.append({"check_name": m.value, **dictm})
            elif isinstance(m, ColumnOp):
                dictm = m.model_dump()
                ser_checks.append({"check_name": m.__class__.__name__, **dictm})
            elif isinstance(m, list):
                ser_checks.append({"check_name": "dummy_list_ops"})
            else:
                raise ValueError(f"Invalid metric: {m}")

        for idx, row in enumerate(data):
            if not req_attrs.issubset(row.keys()):
                raise ValueError(
                    f"Row {idx} is missing required all required attributes for evaluation: {req_attrs}"
                )

        if self.settings.evaluate_locally:
            results = copy.deepcopy(data)
            for idx, check in enumerate(checks):
                if (
                    isinstance(check, ParametricEval)
                    and ser_checks[idx]["check_name"]
                    in PARAMETRIC_EVAL_TO_OPERATOR_MAPPING
                ):
                    # Use the check_name field to get the operator and remove it from ser_checks
                    op = PARAMETRIC_EVAL_TO_OPERATOR_MAPPING[
                        ser_checks[idx].pop("check_name")
                    ](**ser_checks[idx])
                    res = (
                        op.setup(self.settings)
                        .run(pl.DataFrame(data))["output"]
                        .to_dicts()
                    )
                elif isinstance(check, Evals) and check in EVAL_TO_OPERATOR_MAPPING:
                    op = EVAL_TO_OPERATOR_MAPPING[check]
                    op.scenario_description = (
                        scenario_description
                        if not isinstance(scenario_description, list)
                        else scenario_description[idx]
                    )
                    res = (
                        op.setup(self.settings)
                        .run(pl.DataFrame(data))["output"]
                        .to_dicts()
                    )
                elif isinstance(check, ColumnOp):
                    op = Check(name = "dummy", operators = [check])
                    res = (
                        op.setup(self.settings)
                        .run(pl.DataFrame(data))
                    ).to_dicts()
                elif isinstance(check, list):
                    op = Check(name = "dummy", operators = check)
                    res = (
                        op.setup(self.settings)
                        .run(pl.DataFrame(data))
                    ).to_dicts()
                else:
                    res = self.evaluate_on_server(data, [ser_checks[idx]], schema)
                for idx, row in enumerate(res):
                    results[idx].update(row)
        else:
            results = self.evaluate_on_server(data, ser_checks, schema)
        ## local server calls
        try:
            url = "http://localhost:4300/api/public/user"
            client = httpx.Client(
                headers={"uptrain-access-token": "default_key"},
                timeout=httpx.Timeout(7200, connect=5),
            )
            response = client.post(url, json={"name": "default_key"})

            user_id = response.json()["id"]
            checks = []
            for res in results:
                row_check = {}
                for key in res:
                    if key.startswith("score") or key.startswith("explanation"):
                        row_check.update({key: res[key]})
                checks.append(row_check)

            url = "http://localhost:4300/api/public/add_project_data"
            response = client.post(
                url,
                json={
                    "data": results,
                    "checks": checks,
                    "metadata": metadata,
                    "schema_dict": schema.model_dump(),
                    "project": project_name,
                },
            )
        except Exception:
            user_id = "default_key"
            logger.info("Local server not running, start the server to log data and visualize in the dashboard!")
        return results

    def evaluate_on_server(self, data, ser_checks, schema):
        # send in chunks of 50, so the connection doesn't time out waiting for the server
        results = []
        NUM_TRIES, BATCH_SIZE = 3, 50
        for i in range(0, len(data), BATCH_SIZE):
            batch_results = []
            for try_num in range(NUM_TRIES):
                try:
                    logger.info(
                        f"Sending evaluation request for rows {i} to <{i+BATCH_SIZE} to the Uptrain"
                    )
                    batch_results = self.executor.evaluate(
                        data=data[i : i + BATCH_SIZE],
                        checks=ser_checks,
                        metadata={
                            "schema": schema.model_dump(),
                            "uptrain_settings": self.settings.model_dump(),
                        },
                    )
                    break
                except Exception as e:
                    logger.info("Retrying evaluation request")
                    if try_num == NUM_TRIES - 1:
                        logger.error(f"Evaluation failed with error: {e}")
                        raise e

            results.extend(batch_results)
        return results

    def evaluate_experiments(
        self,
        project_name: str,
        data: t.Union[list[dict], pl.DataFrame],
        checks: list[t.Union[str, Evals, ParametricEval]],
        exp_columns: list[str],
        schema: t.Union[DataSchema, dict[str, str], None] = None,
        metadata: t.Optional[dict[str, t.Any]] = None,
    ):
        """Evaluate experiments on the given data.

        Args:
            project_name: Name of the project.
            data: Data to evaluate on. Either a Pandas DataFrame or a list of dicts.
            checks: List of checks to evaluate on.
            exp_columns: List of columns/keys which denote different experiment configurations.
            schema: Schema of the data. Only required if the data attributes aren't typical (question, response, context).
            metadata: Attributes to attach to this dataset. Useful for filtering and grouping in the UI.

        Returns:
            results: List of dictionaries with each data point and corresponding evaluation results for all the experiments.
        """
        if metadata is None:
            metadata = {}

        metadata.update({"uptrain_experiment_columns": exp_columns})

        if schema is None:
            schema = DataSchema()
        elif isinstance(schema, dict):
            schema = DataSchema(**schema)

        results = self.evaluate(
            project_name=project_name,
            data=data,
            checks=checks,
            schema=schema,
            metadata=metadata,
        )

        results = pl.DataFrame(results)
        all_cols = set(results.columns)
        value_cols = list(all_cols - set([schema.question] + exp_columns))
        index_cols = metadata.get("uptrain_index_columns", [schema.question])
        if sum(results.is_duplicated()) > 1:
            logger.info("Duplicates found in data: Removing duplicates")
            results = results.unique()
        exp_results = results.pivot(
                values=value_cols, index=index_cols, columns=exp_columns
            )
        exp_results = exp_results.to_dicts()
        return exp_results

    def evaluate_prompts(
        self,
        project_name: str,
        data: t.Union[list[dict], pl.DataFrame],
        checks: list[t.Union[str, Evals, ParametricEval]],
        prompt: str,
        schema: t.Union[DataSchema, dict[str, str], None] = None,
        metadata: t.Optional[dict[str, t.Any]] = None,
    ):
        """Evaluate experiments on the server and log the results.

        Args:
            project_name: Name of the project.
            data: Data to evaluate on. Either a Pandas DataFrame or a list of dicts.
            checks: List of checks to evaluate on.
            prompt: prompt for generating responses.
            exp_columns: List of columns/keys which denote different experiment configurations.
            schema: Schema of the data. Only required if the data attributes aren't typical (question, response, context).
            metadata: Attributes to attach to this dataset. Useful for filtering and grouping in the UI.

        Returns:
            results: List of dictionaries with each data point and corresponding evaluation results for all the experiments.
        """
        if metadata is None:
            metadata = {}

        base_prompt, prompt_vars = parse_prompt(prompt)

        prompts = []
        context_vars = {}
        context_vars.update(zip(prompt_vars, prompt_vars))
        for idx, item in enumerate(data):
            subs = {k: item[v] for k, v in context_vars.items()}
            msg = base_prompt.format(**subs)
            prompts.append(msg)

        model = metadata["model"]
        dataset = pl.DataFrame(data)
        dataset = dataset.with_columns(
            pl.Series(name="model", values=[model] * len(dataset))
        )
        dataset = dataset.with_columns(pl.Series(name="prompt", values=prompts))

        from uptrain.operators import TextCompletion

        dataset = (
            TextCompletion(
                col_in_prompt="prompt",
                col_in_model="model",
                col_out_completion="response",
                temperature=0.0,
            )
            .setup(self.settings)
            .run(dataset)["output"]
        )

        dataset = dataset.to_dicts()

        if schema is None:
            schema = DataSchema()
        elif isinstance(schema, dict):
            schema = DataSchema(**schema)

        results = self.evaluate(
            project_name=project_name,
            data=dataset,
            checks=checks,
            schema=schema,
            metadata=metadata,
        )
        return results



import os
from uptrain.framework import Settings

SETTINGS = Settings()
SELF_DIR = os.path.dirname(os.path.abspath(__file__))


# uptrain.operators.drift
def test_concept_drift():
    import polars as pl
    from uptrain.operators import ParamsDDM, ConceptDrift
    from uptrain.operators import CsvReader

    # Create an instance of the ParamsDDM class with the parameters
    params_ddm = ParamsDDM(warm_start=500, warning_threshold=2.0, drift_threshold=3.0)

    # Create an instance of the ConceptDrift operator
    op = ConceptDrift(algorithm="DDM", params=params_ddm, col_in_measure="prediction")

    # Run the operator on the input data
    input_data = (
        CsvReader(fpath=os.path.join(SELF_DIR, "data/predictions.csv"))
        .setup(SETTINGS)
        .run()["output"]
    )
    output = op.setup(SETTINGS).run(input_data)["extra"]

    # Check the detected concept drift information
    if output["alert_info"] is not None:
        print("Counter:", output["alert_info"]["counter"])


# uptrain.operators.language.embedding
def test_embedding():
    import polars as pl
    from uptrain.operators import Embedding

    # Create a DataFrame
    df = pl.DataFrame(
        {"text": ["This is the first sentence.", "Here is another sentence."]}
    )

    # Create an instance of the Embedding class
    embedding_op = Embedding(model="MiniLM-L6-v2", col_in_text="text")

    # Generate embeddings for the text column
    embeddings = embedding_op.setup(SETTINGS).run(df)["output"]

    # Print the embeddings
    print(embeddings)


# uptrain.operators.embs
def test_embs_cosine_distribution():
    import polars as pl
    from uptrain.operators import Distribution
    from uptrain.operators import JsonReader

    # Create an instance of the Distribution operator
    op = Distribution(
        kind="cosine_similarity",
        col_in_embs=["context_embeddings", "response_embeddings"],
        col_in_groupby=["question_idx", "experiment_id"],
        col_out=["similarity-context", "similarity-response"],
    )

    # Run the operator on the input data
    input_data = (
        JsonReader(fpath=os.path.join(SELF_DIR, "data/qna_on_docs_samples.jsonl"))
        .setup(SETTINGS)
        .run()["output"]
    )
    output = op.setup(SETTINGS).run(input_data)["output"]

    # Print the output
    print(output)


# uptrain.operators.embs
def test_embs_rouge_score():
    import polars as pl
    from uptrain.operators import Distribution
    from uptrain.operators import JsonReader

    # Create an instance of the Distribution operator
    op = Distribution(
        kind="rouge",
        col_in_embs=["document_text"],
        col_in_groupby=["question_idx", "experiment_id"],
        col_out=["rogue_f1"],
    )
    # Run the operator on the input data
    input_data = (
        JsonReader(fpath=os.path.join(SELF_DIR, "data/qna_on_docs_samples.jsonl"))
        .setup(SETTINGS)
        .run()["output"]
    )
    output = op.setup(SETTINGS).run(input_data)["output"]

    # Print the output
    print(output)


# uptrain.operators.embs
def test_embs_umap_operator():
    import polars as pl
    from uptrain.operators import UMAP
    from uptrain.operators import JsonReader

    # Create an instance of the UMAP operator
    op = UMAP(col_in_embs_1="context_embeddings", col_in_embs_2="response_embeddings")

    # Run the operator on the input data
    input_data = (
        JsonReader(fpath=os.path.join(SELF_DIR, "data/qna_on_docs_samples.jsonl"))
        .setup(SETTINGS)
        .run()["output"]
    )
    output = op.setup(SETTINGS).run(input_data)["output"]

    # Print the output
    print(output)


# uptrain.operators.similarity
def test_cosine_similarity_operator():
    import polars as pl
    import numpy as np
    from uptrain.operators import CosineSimilarity

    # Create a DataFrame
    df = pl.DataFrame(
        {
            "vector1": [np.array([0.1, 0.2, 0.3]), np.array([0.4, 0.5, 0.6])],
            "vector2": [np.array([0.7, 0.8, 0.9]), np.array([0.2, 0.3, 0.4])],
        }
    )

    # Create an instance of the CosineSimilarity class
    similarity_op = CosineSimilarity(
        col_in_vector_1="vector1", col_in_vector_2="vector2"
    )

    # Calculate the cosine similarity between the two vectors
    result = similarity_op.run(df)
    similarity_scores = result["output"]

    # Print the similarity scores
    print(similarity_scores)


# uptrain.operators.metrics
def test_accuracy_operator():
    from uptrain.operators import Accuracy
    from uptrain.operators import CsvReader

    # Create an instance of the Accuracy operator
    op = Accuracy(
        kind="NOT_EQUAL",
        col_in_prediction="prediction",
        col_in_ground_truth="ground_truth",
    )

    # Run the operator on the input data
    input_data = (
        CsvReader(fpath=os.path.join(SELF_DIR, "data/predictions.csv"))
        .setup(SETTINGS)
        .run()["output"]
    )
    accuracy_scores = op.setup(SETTINGS).run(input_data)["output"]

    # Print the accuracy scores
    print(accuracy_scores)


# uptrain.operators.table
def test_table_operator():
    import polars as pl
    from uptrain.operators import ColumnExpand

    # Create a DataFrame
    df = pl.DataFrame({"column1": [1, 2, 3], "column2": ["A", "B", "C"]})

    # Create an instance of the ColumnExpand class
    expand_op = ColumnExpand(
        col_out_names=["column1", "column2"], col_vals=[df["column1"], df["column2"]]
    )

    # Return the input DataFrame as it is
    output_df = expand_op.run(df)["output"]

    # Print the output DataFrame
    print(output_df)


# uptrain.operators.chart
def test_chart_operators():
    import polars as pl
    from uptrain.operators import (
        LineChart,
        ScatterPlot,
        BarChart,
        Histogram,
        MultiPlot,
        CustomPlotlyChart,
    )

    # Create a DataFrame
    df = pl.DataFrame({"x": [1, 2, 3, 4, 5], "y": [10, 20, 15, 25, 30]})

    multiplot = MultiPlot(
        charts=[
            LineChart(x="x", y="y", title="Line Chart", description="Line Chart"),
            ScatterPlot(x="x", y="y", title="Scatter Plot", description="Scatter Plot"),
            BarChart(x="x", y="y", title="Bar Chart", description="Bar Chart"),
            Histogram(x="x", title="Histogram", description="Histogram"),
            CustomPlotlyChart(kind="funnel", x="x", y="y", title="Funnel Chart"),
        ],
        title="MultiPlot",
    )

    # Generate the multiplot
    chart = multiplot.run(df)["extra"]["chart"]

    # Show the chart
    chart.show()


# uptrain.operators.language.rouge
def test_rouge_operator():
    import polars as pl
    from uptrain.operators import RougeScore

    # Create a DataFrame
    df = pl.DataFrame(
        {
            "text_generated": [
                "This is the generated text.",
                "Another generated sentence.",
            ],
            "text_source": [
                "This is the original source text.",
                "This is a different source text.",
            ],
        }
    )

    # Create an instance of the RougeScore class
    rouge_op = RougeScore(score_type="f1")

    # Calculate the Rouge-L scores
    scores = rouge_op.run(df)["output"]

    # Print the Rouge-L scores
    print(scores)


# uptrain.operators.language.bleu
def test_bleu_operator():
    import polars as pl
    from uptrain.operators import BLEUScore

    # Create a DataFrame
    df = pl.DataFrame(
        {
            "text_generated": [
                "This is the generated text.",
                "Another generated sentence.",
            ],
            "text_source": [
                "This is the original source text.",
                "This is a different source text.",
            ],
        }
    )

    # Create an instance of the BleuScore class
    bleu_op = BLEUScore()

    # Calculate the BLEU scores
    scores = bleu_op.run(df)["output"]

    # Print the BLEU scores
    print(scores)


# uptrain.operators.language.meteor
def test_meteor_operator():
    import polars as pl
    from uptrain.operators import METEORScore

    # Create a DataFrame
    df = pl.DataFrame(
        {
            "text_generated": [
                "This is the generated text.",
                "Another generated sentence.",
            ],
            "text_source": [
                "This is the original source text.",
                "This is a different source text.",
            ],
        }
    )

    # Create an instance of the METEORScore class
    meteor_op = METEORScore()

    # Calculate the METEOR scores
    scores = meteor_op.run(df)["output"]

    # Print the METEOR scores
    print(scores)


# uptrain.operators.language.text
def test_docs_link_version_operator():
    import polars as pl
    from uptrain.operators import DocsLinkVersion

    # Create a DataFrame
    df = pl.DataFrame(
        {
            "text": [
                "https://docs.streamlit.io/1.9.0/library/api-reference/charts/st.plotly_chart#stplotly_chart",
                "No version here",
            ]
        }
    )

    # Create an instance of the DocsLinkVersion class
    link_op = DocsLinkVersion(col_in_text="text")

    # Extract the version numbers
    versions = link_op.run(df)["output"]

    # Print the extracted version numbers
    print(versions)


# uptrain.operators.language.text
def test_text_length_operator():
    import polars as pl
    from uptrain.operators import TextLength

    # Create a DataFrame
    df = pl.DataFrame(
        {
            "text": [
                "This is a sample text.",
                "Another example sentence.",
                "Yet another sentence.",
            ]
        }
    )

    # Create an instance of the TextLength class
    length_op = TextLength(col_in_text="text")

    # Calculate the length of each text entry
    lengths = length_op.run(df)["output"]

    # Print the text lengths
    print(lengths)


# uptrain.operators.language.text
def test_text_comparison_operator():
    import polars as pl
    from uptrain.operators import TextComparison

    # Create a DataFrame
    df = pl.DataFrame(
        {
            "text": [
                "This is a sample text.",
                "Another example sentence.",
                "Yet another sentence.",
            ]
        }
    )

    # Set the reference text for comparison
    ref_text = "This is a sample text."

    # Create an instance of the TextComparison class
    comp_op = TextComparison(reference_texts=ref_text, col_in_text="text")

    # Compare each text entry with the reference text
    comparison = comp_op.run(df)["output"]

    # Print the comparison results
    print(comparison)



"""SQLAlchemy models and database session.
"""

import typing as t

from sqlalchemy import (
    create_engine,
    Column,
    String,
    JSON,
    Date,
    DateTime,
    Integer,
    Boolean,
    ForeignKey,
    UniqueConstraint,
)
from sqlalchemy.orm import declarative_base, sessionmaker, scoped_session

from uptrain.utilities.utils import get_uuid, get_current_datetime


SQLBase = declarative_base()


class ModelDataset(SQLBase):
    __tablename__ = "datasets"

    id = Column(String, primary_key=True, nullable=False, default=get_uuid)
    created_at = Column(DateTime, default=get_current_datetime)
    name = Column(String, nullable=False)
    version = Column(Integer, nullable=False)
    address = Column(String, nullable=False)
    user_id = Column(String, ForeignKey("users.id"), nullable=False, index=True)
    rows_count = Column(Integer, nullable=False)

    __table_args__ = (
        UniqueConstraint("user_id", "name", "version", name="uix_dataset"),
    )


class ModelPrompt(SQLBase):
    __tablename__ = "prompts"

    id = Column(String, primary_key=True, nullable=False, default=get_uuid)
    created_at = Column(DateTime, default=get_current_datetime)
    name = Column(String, nullable=False)
    version = Column(Integer, nullable=False)
    prompt = Column(String, nullable=False)
    user_id = Column(String, ForeignKey("users.id"), nullable=False, index=True)

    __table_args__ = (
        UniqueConstraint("user_id", "name", "version", name="uix_prompt"),
    )


class ModelUser(SQLBase):
    __tablename__ = "users"

    id = Column(String, primary_key=True, nullable=False, default=get_uuid)
    created_at = Column(DateTime, default=get_current_datetime)
    updated_at = Column(DateTime, default=get_current_datetime)
    name = Column(String, nullable=False, index=True)
    is_active = Column(Boolean, default=True)

    __table_args__ = (UniqueConstraint("name", name="uix_user"),)


def create_database(db_url):
    """Create the database and tables, and return back a session factory."""
    engine = create_engine(db_url, connect_args={"check_same_thread": False})
    SessionLocal = scoped_session(
        sessionmaker(autocommit=False, autoflush=False, bind=engine)
    )
    SQLBase.metadata.create_all(bind=engine)
    return SessionLocal



from typing import Any

import evals
import evals.metrics
from evals.api import CompletionFn
from evals.prompt.base import is_chat_prompt


class ModelRun(evals.Eval):
    def __init__(
        self,
        completion_fns: list[CompletionFn],
        samples_jsonl: str,
        *args,
        max_tokens: int = 500,
        **kwargs,
    ):
        super().__init__(completion_fns, *args, **kwargs)
        assert len(completion_fns) == 1, "Match only supports one completion fn"
        self.max_tokens = max_tokens
        self.samples_jsonl = samples_jsonl

    def eval_sample(self, sample: Any, *_):
        prompt = sample["input"]
        result = self.completion_fn(
            prompt=prompt,
            temperature=0.0,
        )
        sampled = result.get_completions()[0]

        return evals.record_and_check_match(
            prompt=prompt,
            sampled=sampled,
            expected="",
        )

    def run(self, recorder):
        samples = self.get_samples()
        self.eval_all_samples(recorder, samples)
        model_outputs = []
        for x in recorder._events:
            if "prompt" in x.data:
                model_outputs.append(x.data)
        return model_outputs



"""
Operators for the uptrain.core module. 
"""

from __future__ import annotations
import importlib
import types
import typing as t
import typing_extensions as te

from loguru import logger
from pydantic import ConfigDict, BaseModel
import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings


__all__ = [
    "TYPE_TABLE_OUTPUT",
    "Operator",
    "OpBaseModel",
    "ColumnOp",
    "TransformOp",
    "PlaceholderOp",
    "register_op",
    "deserialize_operator",
    "get_output_col_name_at",
]

# -----------------------------------------------------------
# Base classes for operators
# -----------------------------------------------------------


def make_module_for_custom_ops():
    """Creates a module to hold custom operators. `inspect.getsource` is
    giving me a lot of grief.
    """
    mod = types.ModuleType("_uptrain_custom_ops")
    preamble = [
        "from uptrain.operators import ColumnOp, TransformOp, register_custom_op",
        "import polars as pl",
    ]
    for line in preamble:
        exec(line, mod.__dict__)
    return mod


class TYPE_TABLE_OUTPUT(te.TypedDict):
    output: pl.DataFrame | None
    extra: te.NotRequired[dict]


class Operator(t.Protocol):
    """
    All operator implementations must implement this protocol.

    An operator is usually defined as a pydantic model with the parameters
    necessary to run it. For ex,
    - params necessary to run an algorithm
    - columns required in the input dataset
    - columns that will be generated as output, etc.


    `Setup`: The `setup` method is called before the operator is run. This is where you
    can do any setup work, like loading models, settin up file addresses, etc.

    `Run`: The `run` method is called to run the operator. It takes 0 or more polars
    dataframes as input, and returns a dictionary with the `output` key set to the
    computed dataframe/None. Any extra information can be put in the `extra` key.
    """

    def dict(self) -> dict:
        """
        Serialize this operator to a json dictionary. The objective is to be able to
        recreate the operator from this dict.

        NOTE: If your operator is a pydantic model, the `to_py_types` method in
        `uptrain.utilities` handles this. Though if you have fields with custom
        non-python types, you MUST override and implement both a `dict` and a`from_dict`
        method.
        TODO: With pydantic v2, nested models work with custom (de)serializers. Use that
        when moving.
        """
        ...

    def setup(self, settings: "Settings") -> "Operator":
        """Setup the operator. This must be called before the operator is run."""
        ...

    run: t.Callable[..., TYPE_TABLE_OUTPUT]  # runs the operator


class OpBaseModel(BaseModel):
    """
    Base class you can use if constructing an operator using a pydantic
    model, to get around some of the sharp edges.
    """

    # TODO[pydantic]: The following keys were removed: `smart_union`, `underscore_attrs_are_private`.
    # Check https://docs.pydantic.dev/dev-v2/migration/#changes-to-config for more information.
    model_config = ConfigDict(extra="allow", protected_namespaces=())


class ColumnOp(OpBaseModel):
    """Represents operations that append columns to the input dataset, and
    return it as is.
    """

    def setup(self, settings: "Settings"):
        raise NotImplementedError

    def run(self, *args: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        """
        Runs the operator on the input dataset, and returns it with new columns added.

        Args:
            data (pl.DataFrame): Zero or one dataframe as input.

        Returns:
            A dictionary with the `output` key set to the computed Table. Any extra
                information can be put in the `extra` key.
        """
        raise NotImplementedError


class TransformOp(OpBaseModel):
    """Represents operations that transform the input dataset into another.
    For things like filtering, aggregation, etc.
    """

    def setup(self, settings: Settings):
        raise NotImplementedError

    def run(self, *args: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        """Runs the operator on the input dataset, and returns the transformed dataset.

        Attributes:
            *args (pl.DataFrame): Zero or more dataframes as inputs.

        Returns:
            A dictionary with the `output` key set to the computed dataframe/None. Any extra
                information can be put in the `extra` key.
        """
        raise NotImplementedError


T = t.TypeVar("T")


def register_op(cls: T) -> T:
    """Decorator that marks the class as an Uptrain operator. Useful for (de)serialization."""
    assert isinstance(cls, type), "Only classes can be registered as Uptrain operators"
    assert hasattr(cls, "setup") and hasattr(
        cls, "run"
    ), "All Uptrain operators must define a `setup` and a `run` method."
    op_name = f"{cls.__module__}:{cls.__name__}"
    cls._uptrain_op_name = op_name  # type: ignore
    return cls


def register_custom_op(cls: T) -> T:
    """Decorator that marks the class as a custom Uptrain operator, that is not
    part of the core uptrain package. These are serialized by storing the entire
    source code of the class, and deserialized by exec-ing the source code.

    NOTE: This introduces some restrictions on the operator.
    - This operator is dserialised in an empty namespace, so all imports must
    be done inside the class definition.
    - For typing, use string annotations.
    - Custom operators must be defined in a separate .py file and not in the
    interpreter (like a jupyter notebook).
    """
    cls._uptrain_op_custom = True  # type: ignore
    return register_op(cls)


CUSTOM_OP_MODULE = make_module_for_custom_ops()


def deserialize_operator(data: dict) -> Operator:
    """Deserialize an operator from the serialized dict."""
    op_name = data["op_name"]
    params = data["params"]
    mod_name, cls_name = op_name.split(":")
    try:
        # Check if it is a custom operator, that'd need exec-ing
        if "source" in data:
            exec(data["source"], CUSTOM_OP_MODULE.__dict__)
            klass = getattr(CUSTOM_OP_MODULE, cls_name)
            # to get around serialisation round-trip woes
            klass._uptrain_op_custom_source = data["source"]
            return klass(**params)  # type: ignore
        else:
            ## support for previous operators
            if mod_name.startswith("uptrain.operators"):
                mod_name = "uptrain.operators"
            mod = importlib.import_module(mod_name)
            op = getattr(mod, cls_name)
            # Is it a class implemented by uptrain or a regular pydantic model?
            if hasattr(op, "from_dict"):
                return op.from_dict(params)  # type: ignore
            else:
                return op(**params)  # type: ignore

    except (ModuleNotFoundError, AttributeError) as exc:
        logger.error(
            f"Error when trying to fetch the operator: \n{exc}\n Creating a placeholder op for {op_name}."
        )
        return PlaceholderOp(op_name=op_name, params=data["params"])


# -----------------------------------------------------------
# Utility routines for operators
# -----------------------------------------------------------


def get_output_col_name_at(index: int) -> str:
    return f"_col_{index}"


@register_op
class PlaceholderOp(OpBaseModel):
    """A placeholder operator that does nothing. Used when deserializing an operator,
    that hasn't been registered."""

    op_name: str
    params: dict[str, t.Any]

    def setup(self, settings: "Settings"):
        raise NotImplementedError

    def run(self, *args: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        raise NotImplementedError(
            "This is only a placeholder since the operator: {self.op_name} is not registered with Uptrain. Import the module where it is defined and try again."
        )



from uptrain.utilities.sql_utils import execute_and_compare_sql


# Tests using spider dataset https://yale-lily.github.io/spider
if __name__ == "__main__":
    # TODO: move these to unit tests
    # Issue: Selects country as French instead of France
    predicted_sql = """
    SELECT """
    gt_query = """
    SELECT avg(age) ,  min(age) ,  max(age) FROM singer WHERE country  =  'France'
    """
    res = execute_and_compare_sql(
        predicted_sql,
        gt_query,
        "/Users/bhanu/src/uptrain_experiments/llm/spider/database/concert_singer/concert_singer.sqlite",
    )
    print(res)

    # Issue: selects wrong column
    predicted_sql = """
        SELECT s.Name, s.Song_release_year
FROM singer s
WHERE s.Age = (SELECT MIN(Age) FROM singer)
LIMIT 1;"""
    gt_query = """
        SELECT song_name ,  song_release_year FROM singer ORDER BY age LIMIT 1
        """
    res = execute_and_compare_sql(
        predicted_sql,
        gt_query,
        "/Users/bhanu/src/uptrain_experiments/llm/spider/database/concert_singer/concert_singer.sqlite",
    )
    print(res)

    # Issue: incorrect order of select columns
    predicted_sql = (
        "SELECT City, COUNT(Employee_ID) AS num_employees FROM employee GROUP BY City;"
    )
    gt_query = "SELECT count(*) ,  city FROM employee GROUP BY city"
    res = execute_and_compare_sql(
        predicted_sql,
        gt_query,
        "/Users/bhanu/src/uptrain_experiments/llm/spider/database/employee_hire_evaluation/employee_hire_evaluation.sqlite",
    )
    assert res

    # Ignore column order and row order
    predicted_sql = "SELECT City, COUNT(Employee_ID) AS num_employees FROM employee GROUP BY City order by City;"
    gt_query = "SELECT city, count(*) as cnt FROM employee GROUP BY city order by cnt"
    res = execute_and_compare_sql(
        predicted_sql,
        gt_query,
        "/Users/bhanu/src/uptrain_experiments/llm/spider/database/employee_hire_evaluation/employee_hire_evaluation.sqlite",
    )
    assert res

    # Issue: incorrect GT query
    predicted_sql = (
        "SELECT Name FROM teacher WHERE Hometown <> 'Little Lever Urban District'"
    )
    gt_query = (
        'select name from teacher where hometown != "little lever urban district"'
    )
    res = execute_and_compare_sql(
        predicted_sql,
        gt_query,
        "/Users/bhanu/src/uptrain_experiments/llm/spider/database/course_teach/course_teach.sqlite",
    )
    print(res)

    # Issue: incorrect order of select columns and column case
    predicted_sql = "SELECT PetType, MAX(weight) FROM Pets GROUP BY PetType;"
    gt_query = "SELECT max(weight) ,  petType FROM pets GROUP BY petType"
    res = execute_and_compare_sql(
        predicted_sql,
        gt_query,
        "/Users/bhanu/src/uptrain_experiments/llm/spider/database/pets_1/pets_1.sqlite",
    )
    print(res)



SCENARIO_DESCRIPTION_TEMPLATE = """Here is a brief description of the scenario under which the given response was generated.
{scenario_description}
"""


def parse_scenario_description(scenario_description):
    """
    Parses a scenario description and extracts variables enclosed within '{{' and '}}' or '{' and '}'.

    Args:
        scenario_description (str): The scenario description containing variables.

    Returns:
        tuple: A tuple containing the parsed scenario description and a list of extracted variables.
               The first element is the parsed scenario description with variables formatted as '{var}'.
               The second element is a list of extracted variables.
    """
    scenario_vars = []
    if scenario_description is not None and len(scenario_description):
        scenario_description = SCENARIO_DESCRIPTION_TEMPLATE.format(
            scenario_description=scenario_description
        )
        if "{{" in scenario_description:
            scenario_vars = [
                x.split("}}")[0] for x in scenario_description.split("{{")[1:]
            ]
            for var in scenario_vars:
                scenario_description = scenario_description.replace(
                    "{{" + var + "}}", "{" + var + "}"
                )
        elif "{" in scenario_description:
            scenario_vars = [
                x.split("}")[0] for x in scenario_description.split("{")[1:]
            ]
    else:
        scenario_description = ""
    return scenario_description, scenario_vars



"""This module hosts built-in checks for some common LLM evaluation tasks."""

from __future__ import annotations
from .checks import Check
from uptrain.operators import Histogram
from uptrain.operators import (
    ResponseFactualScore,
    ValidResponseScore,
    ResponseCompleteness,
    ResponseRelevance,
    ResponseConsistency,
    ResponseConciseness,
    ContextRelevance,
    PromptInjectionScore,
    LanguageCritique,
    ToneCritique,
    ResponseCompletenessWrtContext,
    GuidelineAdherenceScore,
    ResponseMatchingScore,
    ConversationSatisfactionScore,
    CodeHallucinationScore,
    JailbreakDetectionScore,
    SubQueryCompleteness,
    ContextReranking,
    ContextConciseness,
    MultiQueryAccuracy,
)


# -----------------------------------------------------------
# Response Quality
# -----------------------------------------------------------


def CheckResponseCompleteness():
    return Check(
        name="response_completeness_score",
        operators=[ResponseCompleteness()],
        plots=[Histogram(x="score_response_completeness")],
    )


def CheckResponseConciseness():
    return Check(
        name="response_conciseness_score",
        operators=[ResponseConciseness()],
        plots=[Histogram(x="score_response_conciseness")],
    )


def CheckResponseRelevance():
    return Check(
        name="response_relevance_score",
        operators=[ResponseRelevance()],
        plots=[Histogram(x="score_response_relevance")],
    )


def CheckValidResponse():
    return Check(
        name="valid_response_score",
        operators=[ValidResponseScore()],
        plots=[Histogram(x="score_valid_response")],
    )


def CheckResponseConsistency():
    return Check(
        name="response_consistency_score",
        operators=[ResponseConsistency()],
        plots=[Histogram(x="score_response_consistency")],
    )


def CheckResponseMatching(method="llm"):
    return Check(
        name="response_matching_score",
        operators=[ResponseMatchingScore(method=method)],
        plots=[Histogram(x="score_response_match")],
    )


# -----------------------------------------------------------
# Context Quality
# -----------------------------------------------------------


def CheckContextRelevance():
    return Check(
        name="score_context_relevance",
        operators=[ContextRelevance()],
        plots=[Histogram(x="score_context_relevance")],
    )


def CheckResponseCompletenessWrtContext():
    return Check(
        name="response_completeness_wrt_context_score",
        operators=[ResponseCompletenessWrtContext()],
        plots=[Histogram(x="score_response_completeness_wrt_context")],
    )


def CheckResponseFacts():
    return Check(
        name="score_factual_accuracy",
        operators=[ResponseFactualScore()],
        plots=[Histogram(x="score_factual_accuracy")],
    )


def CheckContextReranking():
    return Check(
        name="context_reranking_score",
        operators=[ContextReranking()],
        plots=[Histogram(x="score_context_reranking")],
    )


def CheckContextConciseness():
    return Check(
        name="context_conciseness_score",
        operators=[ContextConciseness()],
        plots=[Histogram(x="score_context_conciseness")],
    )


# -----------------------------------------------------------
# Language Proficiency
# -----------------------------------------------------------


def CheckLanguageQuality():
    return Check(
        name="language_critique_score",
        operators=[LanguageCritique()],
        plots=[
            Histogram(x="score_language_critique"),
        ],
    )


def CheckToneQuality(llm_persona):
    return Check(
        name="tone_critique_score",
        operators=[ToneCritique(llm_persona=llm_persona)],
        plots=[Histogram(x="score_critique_tone")],
    )


# -----------------------------------------------------------
# Code generation
# -----------------------------------------------------------


def CheckCodeHallucination():
    return Check(
        name="code_hallucination_score",
        operators=[CodeHallucinationScore()],
        plots=[Histogram(x="score_code_hallucination")],
    )


# -----------------------------------------------------------
# Conversation Quality
# -----------------------------------------------------------


def CheckConversationSatisfaction(user_role="user", llm_role="assistant"):
    return Check(
        name="conversation_satisfaction_score",
        operators=[
            ConversationSatisfactionScore(user_role=user_role, llm_role=llm_role)
        ],
        plots=[Histogram(x="score_conversation_satisfaction")],
    )


# -----------------------------------------------------------
# Custom Evaluations
# -----------------------------------------------------------


def CheckGuidelineAdherence(
    guideline, guideline_name="guideline", response_schema=None
):
    return Check(
        name=f"{guideline_name}_adherence_score",
        operators=[
            GuidelineAdherenceScore(
                guideline=guideline,
                guideline_name=guideline_name,
                response_schema=response_schema,
            )
        ],
        plots=[Histogram(x=f"score_{guideline_name}_adherence")],
    )


# -----------------------------------------------------------
# Security
# -----------------------------------------------------------


def CheckPromptInjection():
    return Check(
        name="prompt_injection_score",
        operators=[PromptInjectionScore()],
        plots=[Histogram(x="score_prompt_injection")],
    )


def CheckJailbreakDetection():
    return Check(
        name="j`ailbreak_detection_score",
        operators=[JailbreakDetectionScore()],
        plots=[Histogram(x="score_jailbreak_attempted")],
    )


# -----------------------------------------------------------
# Subquery
# -----------------------------------------------------------


def CheckSubQueryCompleteness():
    return Check(
        name="sub_query_completeness_score",
        operators=[SubQueryCompleteness()],
        plots=[Histogram(x="score_sub_query_completeness")],
    )


# -----------------------------------------------------------
# MultiQuery
# -----------------------------------------------------------

def CheckMultiQueryAccuracy():
    return Check(
        name="multi_query_accuracy_score",
        operators=[MultiQueryAccuracy()],
        plots=[Histogram(x="score_multi_query_accuracy")],
    )


"""
Implement checks to evaluate the tonality of response. 

This module provides the `ToneCritique` class, which evaluates if the tone of the given response matches with the desired tone. It provides a score for each of the 
aspects on a scale of 0 to 1, along with an explanation for the score. 
"""

from __future__ import annotations
import json
import typing as t

from loguru import logger
import polars as pl

from uptrain.operators.language.llm import LLMMulticlient
from uptrain.operators.language.prompts.classic import (
    CRITIQUE_TONE_PROMPT_TEMPLATE,
)
from uptrain.operators.language.prompts.few_shots import (
    CRITIQUE_TONE_FEW_SHOT__CLASSIFY,
    CRITIQUE_TONE_FEW_SHOT__COT,
)
from uptrain.operators.language.prompts.instructions import CHAIN_OF_THOUGHT, CLASSIFY
from uptrain.operators.language.prompts.output_format import (
    CRITIQUE_TONE_OUTPUT_FORMAT__CLASSIFY,
    CRITIQUE_TONE_OUTPUT_FORMAT__COT,
)
from uptrain.utilities.prompt_utils import parse_scenario_description

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import (
    register_op,
    ColumnOp,
    TYPE_TABLE_OUTPUT,
)

from uptrain.utilities import polars_to_json_serializable_dict


@register_op
class ToneCritique(ColumnOp):
    """
    Operator to assess the tone of machine generated responses.

    Attributes:
        llm_persona (str): The persona the chatbot being assessed was expected to follow
        col_response (str): The name of the input column containing response text
        col_out (str): The name of the output column containing the scores
        scenario_description (str | None): Optional scenario description to incorporate in the evaluation prompt
        score_mapping (dict): Mapping of different grades to float scores

    Raises:
        Exception: Raises exception for any failed evaluation attempts

    """

    llm_persona: str = "helpful-chatbot"
    col_response: str = "response"
    col_out: str = "score_critique_tone"
    scenario_description: t.Optional[str] = None
    score_mapping: dict = {"A": 1.0, "B": 0.5, "C": 0.0}

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self.settings = settings
        if self.settings.evaluate_locally and (
            self.settings.uptrain_access_token is None
            or not len(self.settings.uptrain_access_token)
        ):
            self._api_client = LLMMulticlient(settings)
        else:
            self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = polars_to_json_serializable_dict(data)
        for row in data_send:
            row["response"] = row.pop(self.col_response)

        try:
            if self.settings.evaluate_locally and (
                self.settings.uptrain_access_token is None
                or not len(self.settings.uptrain_access_token)
            ):
                results = self.evaluate_local(data_send)
            else:
                results = self._api_client.evaluate(
                    "critique_tone", data_send, {"llm_persona": self.llm_persona}
                )
        except Exception as e:
            logger.error(f"Failed to run evaluation for `ToneCritique`: {e}")
            raise e

        assert results is not None
        return {
            "output": data.with_columns(
                pl.from_dicts(results).rename({"score_critique_tone": self.col_out})
            )
        }

    def critique_tone_classify_validate_func(self, llm_output):
        is_correct = True
        is_correct = is_correct and ("Choice" in llm_output)
        is_correct = is_correct and llm_output["Choice"] in ["A", "B", "C"]
        return is_correct

    def critique_tone_cot_validate_func(self, llm_output):
        is_correct = self.critique_tone_classify_validate_func(llm_output)
        is_correct = is_correct and ("Reasoning" in llm_output)
        return is_correct

    def evaluate_local(self, data):
        """
        Our methodology is based on the model grade evaluation introduced by openai evals.
        """

        self.scenario_description, scenario_vars = parse_scenario_description(
            self.scenario_description
        )
        input_payloads = []
        if self.settings.eval_type == "basic":
            few_shot_examples = CRITIQUE_TONE_FEW_SHOT__CLASSIFY
            output_format = CRITIQUE_TONE_OUTPUT_FORMAT__CLASSIFY
            validation_func = self.critique_tone_classify_validate_func
            prompting_instructions = CLASSIFY
        elif self.settings.eval_type == "cot":
            few_shot_examples = CRITIQUE_TONE_FEW_SHOT__COT
            output_format = CRITIQUE_TONE_OUTPUT_FORMAT__COT
            validation_func = self.critique_tone_cot_validate_func
            prompting_instructions = CHAIN_OF_THOUGHT
        else:
            raise ValueError(
                f"Invalid eval_type: {self.settings.eval_type}. Must be either 'basic' or 'cot'"
            )

        for idx, row in enumerate(data):
            kwargs = row
            kwargs.update(
                {
                    "output_format": output_format,
                    "prompting_instructions": prompting_instructions,
                    "few_shot_examples": few_shot_examples,
                    "llm_persona": self.llm_persona,
                }
            )
            try:
                grading_prompt_template = CRITIQUE_TONE_PROMPT_TEMPLATE.replace(
                    "{scenario_description}", self.scenario_description
                ).format(**kwargs)
            except KeyError as e:
                raise KeyError(
                    f"Missing required attribute(s) for scenario description: {e}"
                )
            input_payloads.append(
                self._api_client.make_payload(idx, grading_prompt_template)
            )
        output_payloads = self._api_client.fetch_responses(
            input_payloads, validation_func
        )

        results = []

        for res in output_payloads:
            idx = res.metadata["index"]
            output = {"score_critique_tone": None, "explanation_critique_tone": None}
            try:
                score = self.score_mapping[
                    json.loads(res.response.choices[0].message.content)["Choice"]
                ]
                output["score_critique_tone"] = float(score)
                output["explanation_critique_tone"] = res.response.choices[
                    0
                ].message.content
            except Exception:
                logger.error(
                    f"Error when processing payload at index {idx}: {res.error}"
                )
            results.append((idx, output))

        results = [val for _, val in sorted(results, key=lambda x: x[0])]
        return results



"""
Implement operators to do vector-search and retrieval documents for a given user query. 
"""

from __future__ import annotations
import typing as t
import copy

import numpy as np
import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import (
    TransformOp,
    register_op,
    TYPE_TABLE_OUTPUT,
)
from uptrain.operators.embedding.embedding import Embedding
from uptrain.operators.io.base import JsonReader, CsvReader
from uptrain.utilities import lazy_load_dep, polars_to_pandas

faiss = lazy_load_dep("faiss", "faiss")


@register_op
class VectorSearch(TransformOp):
    """

     Attributes:
        col_in_query (str): Column name for the input query to be used for retrieval
        colk_in_document (str): Column name for the input document to be used for retrieval
        col_out (str): Column name for the retrieved_context
        documents (t.Union[list[str], str, JsonReader, CsvReader]): List of documents or path to the file containing the documents
        embeddings_model (str): Name of the embeddings model
        top_k (int): Top K documents will be retrieved.
        distance_metric (str): One of ['cosine_similarity', 'l2_distance']
        embedding_batch_size (int): Batch size for the embeddings model

    Raises:
        Exception: Raises exception for any failed evaluation attempts

    """

    col_in_query: str = "question"
    col_in_document: str = ""
    col_out: str = "context"
    documents: t.Union[list[str], str, JsonReader, CsvReader] = None
    embeddings_model: str = ""
    top_k: int = 1
    distance_metric: str = t.Literal["cosine_similarity", "l2_distance"]
    embedding_batch_size: int = 128

    def setup(self, settings: t.Optional[Settings] = None):
        read_op = None
        if isinstance(self.documents, list):
            documents_table = pl.DataFrame({"document": self.documents})
        elif isinstance(self.documents, str):
            if self.documents[-5:] == ".json":
                read_op = JsonReader(fpath=self.documents)
            elif self.documents[-6:] == ".jsonl":
                read_op = JsonReader(fpath=self.documents)
            elif self.documents[-4:] == ".csv":
                read_op = CsvReader(fpath=self.documents)
            else:
                read_op = JsonReader(fpath=self.documents)
                # raise Exception("File formats other than jsonl and csv are not supported")
        elif isinstance(self.documents, JsonReader) or isinstance(
            self.documents, CsvReader
        ):
            read_op = self.documents
        else:
            raise Exception(f"{type(self.documents)} is not supported")

        if read_op is not None:
            documents_table = pl.DataFrame(
                {
                    "document": read_op.setup(settings).run()["output"][
                        self.col_in_document
                    ]
                }
            )

        documents_table = pl.DataFrame(polars_to_pandas(documents_table))

        emb_op = Embedding(
            model=self.embeddings_model,
            col_in_text="document",
            col_out="document_embeddings",
            batch_size=self.embedding_batch_size,
        )
        doc_embeddings = np.array(
            list(
                emb_op.setup(settings).run(documents_table)["output"][
                    "document_embeddings"
                ]
            )
        )

        self.documents_list = documents_table["document"]

        if self.distance_metric == "cosine_similarity":
            self.vectorstore = faiss.IndexFlatIP(len(doc_embeddings[0]))
        elif self.distance_metric == "l2_distance":
            self.vectorstore = faiss.IndexFlatL2(len(doc_embeddings[0]))
        else:
            raise Exception(f"{self.distance_metric} is not allowed")
        self.vectorstore.add(doc_embeddings)

        self._settings = settings
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        emb_op = Embedding(
            model=self.embeddings_model,
            col_in_text=self.col_in_query,
            col_out=self.col_in_query + "_embeddings",
        )
        query_emb_res = emb_op.setup(self._settings).run(data)["output"].to_dicts()

        res = []
        for emb_res in query_emb_res:
            this_retrieved_scores, this_retrieved_idxs = self.vectorstore.search(
                np.array([emb_res[self.col_in_query + "_embeddings"]]), self.top_k
            )
            this_retrieved_documents = [
                self.documents_list[int(x)] for x in this_retrieved_idxs[0]
            ]

            del emb_res[self.col_in_query + "_embeddings"]

            this_retriev_res = copy.deepcopy(emb_res)
            this_retriev_res = []
            for _ in range(len(this_retrieved_documents)):
                this_retriev_res.append(copy.deepcopy(emb_res))
            for jdx, row in enumerate(this_retriev_res):
                row.update(
                    {
                        self.col_out: this_retrieved_documents[jdx],
                        "retrieval_rank": jdx + 1,
                        "retrieval_similarity_score": this_retrieved_scores[0][jdx],
                    }
                )
            res.extend(this_retriev_res)

        return {"output": pl.DataFrame(res)}



"""
Implement operators to evaluate conversation datapoints for a 
LLM-powered agent pipeline. 
"""

from __future__ import annotations
import json
import typing as t

from loguru import logger
import polars as pl

from uptrain.operators.language.llm import LLMMulticlient
from uptrain.operators.language.prompts.classic import (
    CONVERSATION_SATISFACTION_PROMPT_TEMPLATE,
)
from uptrain.operators.language.prompts.few_shots import (
    CONVERSATION_SATISFACTION_FEW_SHOT__CLASSIFY,
    CONVERSATION_SATISFACTION_FEW_SHOT__COT,
)
from uptrain.operators.language.prompts.instructions import CHAIN_OF_THOUGHT, CLASSIFY
from uptrain.operators.language.prompts.output_format import (
    CONVERSATION_SATISFACTION_OUTPUT_FORMAT__CLASSIFY,
    CONVERSATION_SATISFACTION_OUTPUT_FORMAT__COT,
)
from uptrain.utilities.prompt_utils import parse_scenario_description

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import (
    register_op,
    ColumnOp,
    TYPE_TABLE_OUTPUT,
)

from uptrain.utilities import polars_to_json_serializable_dict


@register_op
class ConversationSatisfactionScore(ColumnOp):
    """
    Grade if the user is satifised with the conversation with the ChatBot.

     Attributes:
        col_conversation (str): Column name for the stored conversations.
        col_out (str): Column name to output scores
        assistant_persona (str): The personality description of the assistant.
        user_role (str): Designation of the user in the conversation.
        assistant_role (str): Designation of the assistant in the conversation.
        scenario_description (str | None): Optional scenario description to incorporate in the evaluation prompt.

    Raises:
        Exception: Raises exception for any failed evaluation attempts

    """

    col_conversation: str = "conversation"
    col_out: str = "score_conversation_satisfaction"
    assistant_persona: t.Union[str, None] = None
    user_role: str = "User"
    assistant_role: str = "Assistant"
    scenario_description: t.Optional[str] = None
    score_mapping: dict = {"A": 0.0, "B": 0.5, "C": 1.0}

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self.settings = settings
        if self.settings.evaluate_locally and (
            self.settings.uptrain_access_token is None
            or not len(self.settings.uptrain_access_token)
        ):
            self._api_client = LLMMulticlient(settings)
        else:
            self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = polars_to_json_serializable_dict(data)
        for row in data_send:
            row["conversation"] = row[self.col_conversation]

        try:
            if self.settings.evaluate_locally and (
                self.settings.uptrain_access_token is None
                or not len(self.settings.uptrain_access_token)
            ):
                results = self.evaluate_local(data_send)
            else:
                results = self._api_client.evaluate(
                    "ConversationSatisfaction",
                    data_send,
                    {
                        "user_persona": self.user_role,
                        "llm_persona": self.assistant_role,
                    },
                )

        except Exception as e:
            logger.error(
                f"Failed to run evaluation for `ConversationSatisfactionScore`: {e}"
            )
            raise e

        assert results is not None
        return {
            "output": data.with_columns(
                pl.from_dicts(results).rename(
                    {"score_conversation_satisfaction": self.col_out}
                )
            )
        }

    def conversation_satisfaction_classify_validate_func(self, llm_output):
        is_correct = True
        is_correct = is_correct and ("Choice" in llm_output)
        is_correct = is_correct and llm_output["Choice"] in ["A", "B", "C"]
        return is_correct

    def conversation_satisfaction_cot_validate_func(self, llm_output):
        is_correct = self.conversation_satisfaction_classify_validate_func(llm_output)
        is_correct = is_correct and ("Reasoning" in llm_output)
        return is_correct

    def evaluate_local(self, data):
        """
        Our methodology is based on the model grade evaluation introduced by openai evals.
        """

        self.scenario_description, scenario_vars = parse_scenario_description(
            self.scenario_description
        )
        input_payloads = []
        if self.settings.eval_type == "basic":
            few_shot_examples = CONVERSATION_SATISFACTION_FEW_SHOT__CLASSIFY
            output_format = CONVERSATION_SATISFACTION_OUTPUT_FORMAT__CLASSIFY
            validation_func = self.conversation_satisfaction_classify_validate_func
            prompting_instructions = CLASSIFY
        elif self.settings.eval_type == "cot":
            few_shot_examples = CONVERSATION_SATISFACTION_FEW_SHOT__COT
            output_format = CONVERSATION_SATISFACTION_OUTPUT_FORMAT__COT
            validation_func = self.conversation_satisfaction_cot_validate_func
            prompting_instructions = CHAIN_OF_THOUGHT
        else:
            raise ValueError(
                f"Invalid eval_type: {self.settings.eval_type}. Must be either 'basic' or 'cot'"
            )

        for idx, row in enumerate(data):
            kwargs = row
            kwargs.update(
                {
                    "output_format": output_format,
                    "prompting_instructions": prompting_instructions,
                    "few_shot_examples": few_shot_examples,
                    "user_role": self.user_role,
                    "assistant_role": self.assistant_role,
                    "assistant_persona": self.assistant_persona,
                }
            )
            try:
                grading_prompt_template = (
                    CONVERSATION_SATISFACTION_PROMPT_TEMPLATE.replace(
                        "{scenario_description}", self.scenario_description
                    ).format(**kwargs)
                )
            except KeyError as e:
                raise KeyError(
                    f"Missing required attribute(s) for scenario description: {e}"
                )
            input_payloads.append(
                self._api_client.make_payload(idx, grading_prompt_template)
            )
        output_payloads = self._api_client.fetch_responses(
            input_payloads, validation_func
        )

        results = []
        for res in output_payloads:
            idx = res.metadata["index"]
            output = {
                "score_conversation_satisfaction": None,
                "explanation_conversation_satisfaction": None,
            }
            try:
                score = self.score_mapping[
                    json.loads(res.response.choices[0].message.content)["Choice"]
                ]
                output["score_conversation_satisfaction"] = float(score)
                output["explanation_conversation_satisfaction"] = res.response.choices[
                    0
                ].message.content
            except Exception:
                logger.error(
                    f"Error when processing payload at index {idx}: {res.error}"
                )
            results.append((idx, output))
        results = [val for _, val in sorted(results, key=lambda x: x[0])]

        return results



__all__ = [
    "builtins",
    "Check",
    "CheckSet",
    "ExperimentArgs",
    "EvalLLM",
    "OperatorDAG",
    "Settings",
    "Signal",
    "APIClient",
    "DataSchema",
    "Evals",
    "CritiqueTone",
    "GuidelineAdherence",
    "ResponseMatching",
    "ConversationSatisfaction",
    "CustomPromptEval",
    "RcaTemplate",
    "JailbreakDetection",
]

from .base import OperatorDAG, Settings
from .checks import CheckSet, Check, ExperimentArgs
from .signal import Signal
from .remote import APIClient, DataSchema
from .builtins import builtins
from .evals import (
    Evals,
    CritiqueTone,
    GuidelineAdherence,
    ResponseMatching,
    ConversationSatisfaction,
    CustomPromptEval,
    JailbreakDetection,
)
from .rca_templates import RcaTemplate
from .evalllm import EvalLLM
